"""Tests for the bidsschematools package."""

from bidsschematools.render import tables


def test_make_entity_table(schema_obj):
    """
    Test whether expected entities are present and listed correctly.
    This should be robust with respect to schema format.
    """
    entity_table = tables.make_entity_table(schema_obj)

    # Non-exhaustive list covering both value and index formats
    expected_entities = [
        "[`acq-<label>`](./appendices/entities.md#acq)",
        "[`ses-<label>`](./appendices/entities.md#ses)",
        "[`sample-<label>`](./appendices/entities.md#sample)",
        "[`task-<label>`](./appendices/entities.md#task)",
        "[`acq-<label>`](./appendices/entities.md#acq)",
        "[`ce-<label>`](./appendices/entities.md#ce)",
        "[`trc-<label>`](./appendices/entities.md#trc)",
        "[`stain-<label>`](./appendices/entities.md#stain)",
        "[`rec-<label>`](./appendices/entities.md#rec)",
        "[`dir-<label>`](./appendices/entities.md#dir)",
        "[`run-<index>`](./appendices/entities.md#run)",
    ]

    for expected_entity in expected_entities:
        assert expected_entity in entity_table


def test_make_suffix_table(schema_obj):
    """
    Test whether expected suffixes are present and listed with correct names.
    Values are hard-coded from the present YAML, but should nevertheless be robust
    with respect to schema format, other than case changes for the names.
    """
    target_suffixes = [
        "beh",
        "cbv",
        "dwi",
    ]
    suffix_table = tables.make_suffix_table(schema_obj, target_suffixes)

    expected_names = [
        "Behavioral recording",
        "Cerebral blood volume image",
        "Diffusion-weighted image",
    ]

    for expected_name in expected_names:
        assert expected_name in suffix_table


def test_make_sidecar_table(schema_obj):
    """
    Test whether expected metadata fields are present and the requirement level is
    applied correctly.
    This should be robust with respect to schema format.
    """
    # mri.MRISpatialEncoding selected for having some level and description addenda
    rendered_table = tables.make_sidecar_table(schema_obj, "mri.MRISpatialEncoding").split("\n")

    assert rendered_table[0].startswith("| **Key name**")
    assert rendered_table[1].startswith("|-------------")

    fields = schema_obj.rules.sidecars.mri.MRISpatialEncoding.fields
    assert len(rendered_table) == len(fields) + 2

    for field, render_row in zip(fields, rendered_table[2:]):
        assert render_row.startswith(f"| [{field}](")
        spec = fields[field]
        if isinstance(spec, str):
            level = spec
            level_addendum = ""
            description_addendum = ""
        else:
            level = spec["level"]
            level_addendum = spec.get("level_addendum", "").replace("required", "REQUIRED")
            description_addendum = spec.get("description_addendum", "")

        assert level.upper() in render_row
        assert level_addendum.split("\n")[0] in render_row
        assert description_addendum.split("\n")[0] in render_row


def test_make_metadata_table(schema_obj):
    """
    Test whether expected metadata fields are present and the requirement level is
    applied correctly.
    This should be robust with respect to schema format.
    """
    target_metadata = {
        "Authors": "required",
        "BIDSVersion": "required",
        "DatasetDOI": "optional",
    }
    metadata_table = tables.make_metadata_table(schema_obj, target_metadata).split("\n")

    metadata_tracking = list(target_metadata.keys())

    for line in metadata_table:
        for i in metadata_tracking:
            if i in line:
                # Is the requirement level displayed correctly?
                assert target_metadata[i].upper() in line
                # Mark found
                metadata_tracking.remove(i)

    # Have we found all fields?
    assert len(metadata_tracking) == 0


def test_make_columns_table(schema_obj):
    """
    Test whether expected columns are present and the requirement level is
    applied correctly.
    This should be robust with respect to schema format.
    """
    # mri.MRISpatialEncoding selected for having some level and description addenda
    rendered_table = tables.make_columns_table(
        schema_obj,
        "modality_agnostic.Participants",
    ).split("\n")

    assert rendered_table[0].startswith("| **Column name**")
    assert rendered_table[1].startswith("|----------------")

    fields = schema_obj.rules.tabular_data.modality_agnostic.Participants.columns
    assert len(rendered_table) == len(fields) + 3  # header + orientation + add. cols. row

    for field, render_row in zip(fields, rendered_table[2:-1]):
        assert render_row.startswith(f"| [{field}](")
        spec = fields[field]
        if isinstance(spec, str):
            level = spec
            level_addendum = ""
            description_addendum = ""
        else:
            level = spec["level"]
            level_addendum = spec.get("level_addendum", "").replace("required", "REQUIRED")
            description_addendum = spec.get("description_addendum", "")

        assert level.upper() in render_row
        assert level_addendum.split("\n")[0] in render_row
        assert description_addendum.split("\n")[0] in render_row



import json
import logging
import os
import re
import sys
from itertools import chain

import click

from .rules import regexify_filename_rules
from .schema import load_schema
from .utils import configure_logger, get_logger
from .validator import _bidsignore_check

lgr = get_logger()


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet", count=True)
def cli(verbose, quiet):
    """BIDS Schema Tools"""
    verbose = verbose - quiet
    configure_logger(get_logger(level=logging.WARNING - verbose * 10))


@cli.command()
@click.option("--schema")
@click.option("--output", default="-")
@click.pass_context
def export(ctx, schema, output):
    """Export BIDS schema to JSON document"""
    schema = load_schema(schema)
    text = schema.to_json()
    if output == "-":
        lgr.debug("Writing to stdout")
        print(text)
    else:
        output = os.path.abspath(output)
        lgr.debug(f"Writing to {output}")
        with open(output, "w") as fobj:
            fobj.write(text)


@cli.command()
@click.option("--output", default="-")
@click.pass_context
def export_metaschema(ctx, output):
    """Export BIDS schema to JSON document"""
    from .data import load

    metaschema = load.readable("metaschema.json").read_text()
    if output == "-":
        print(metaschema, end="")
    else:
        output = os.path.abspath(output)
        with open(output, "w") as fobj:
            fobj.write(metaschema)


@cli.command("pre-receive-hook")
@click.option("--schema", "-s", type=click.Path(), help="Path to the BIDS schema")
@click.option(
    "--input", "-i", "input_", default="-", type=click.Path(), help="Input file (default: stdin)"
)
@click.option(
    "--output",
    "-o",
    "output",
    default="-",
    type=click.Path(),
    help="Output file (default: stdout)",
)
def pre_receive_hook(schema, input_, output):
    """Validate filenames from a list of files against the BIDS schema

    The expected input takes the following form:

    ```
    bids-hook-v2
    {"Name": "My dataset", "BIDSVersion": "1.9.0", "DatasetType": "raw"}
    ignore-pattern1
    ...
    ignore-patternN
    0001
    .datalad/config
    .gitattributes
    CHANGES
    README
    dataset_description.json
    participants.tsv
    sub-01/anat/sub-01_T1w.nii.gz
    ...
    ```

    The header identifies the protocol version. For protocol ``bids-hook-v2``,
    the second line MUST be the dataset_description.json file, with any newlines removed.
    The following lines, up to the line containing "0001", are ignore patterns
    from the .bidsignore file. The lines following "0001" are the filenames to
    be validated.

    This is intended to be used in a git pre-receive hook.
    """
    schema = load_schema(schema)

    # Slurp inputs for now; we can think about streaming later
    if input_ == "-":
        stream = sys.stdin
    else:
        stream = open(input_)

    first_line = next(stream)
    if first_line == "bids-hook-v2\n":
        # V2 format: header line, description JSON, followed by legacy format
        description_str = next(stream)
        fail = False
        try:
            description: dict = json.loads(description_str)
        except json.JSONDecodeError:
            fail = True
        if fail or not isinstance(description, dict):
            lgr.critical("Protocol error: invalid JSON in description")
            lgr.critical("Dataset description must be one JSON object, written to a single line")
            lgr.critical("Received: %s", description_str)
            stream.close()
            sys.exit(2)
    else:
        # Legacy: ignore patterns, followed by "0001", followed by filenames
        stream = chain([first_line], stream)
        description = {}

    dataset_type = description.get("DatasetType", "raw")
    lgr.info("Dataset type: %s", dataset_type)

    ignore = []
    for line in stream:
        if line == "0001\n":
            break
        ignore.append(line.strip())
    lgr.info("Ignore patterns found: %d", len(ignore))

    all_rules = chain.from_iterable(
        regexify_filename_rules(group, schema, level=2)
        for group in (schema.rules.files.common, schema.rules.files.raw)
    )
    if dataset_type == "derivative":
        all_rules = chain(
            all_rules,
            regexify_filename_rules(schema.rules.files.derivatives, schema, level=2),
        )

    regexes = [rule["regex"] for rule in all_rules]
    # XXX Hack for phenotype files - this can be removed once we
    # have a schema definition for them
    regexes.append(r"phenotype/.*\.(tsv|json)")

    output = sys.stdout if output == "-" else open(output, "w")

    rc = 0
    any_files = False
    valid_files = 0
    with output:
        for filename in stream:
            if not any_files:
                lgr.debug("Validating files, first file: %s", filename)
                any_files = True
            filename = filename.strip()
            if any(_bidsignore_check(pattern, filename, "") for pattern in ignore):
                continue
            if not any(re.match(regex, filename) for regex in regexes):
                print(filename, file=output)
                rc = 1
            else:
                valid_files += 1

    if valid_files == 0:
        lgr.error("No files to validate")
        rc = 2

    stream.close()
    sys.exit(rc)


if __name__ == "__main__":
    cli()



"""BIDS validation context code generator

This module produces source code for BIDS validation context classes.
The source code can be executed at import time to ensure that classes
are consistent with the current state of the BIDS schema.

When packaging, calling modules should be replaced with the source code
they execute, ensuring that the classes are available for static analysis
and can be compiled to bytecode.

This module can be removed during packaging.
"""

from __future__ import annotations

import json
from textwrap import dedent, indent

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any, Callable, Protocol

    class Spec(Protocol):
        prelude: str
        class_def: str
        attr_def: str
        proto_prefix: str


def with_indent(spaces: int, /) -> Callable[[str], str]:
    def decorator(attr: str) -> str:
        return indent(dedent(attr), " " * spaces)

    return decorator


class ProtocolSpec:
    prelude = dedent('''\
    """BIDS validation context definitions

    The classes in this module are used to define the context for BIDS validation.
    The context is a namespace that contains relevant information about the dataset
    as a whole and an individual file to be validated.

    These classes are used to define the structure of the context,
    but they cannot be instantiated directly.
    Conforming subtypes need only match the structure of these classes,
    and do not need to inherit from them.
    It is recommended to import this module in an ``if TYPE_CHECKING`` block
    to avoid import costs.

    The classes use ``@property`` decorators to indicate that subtypes need only
    provide read access to the attributes, and may restrict writing, for example,
    when calculating attributes dynamically based on other attributes.

    Note that some type checkers will not match classes that use
    :class:`functools.cached_property`.
    To permit this, add the following to your module::

        if TYPE_CHECKING:
            cached_property = property
        else:
            from functools import cached_property

    This module has been auto-generated from the BIDS schema version {version}.
    """

    from __future__ import annotations

    from collections.abc import Mapping, Sequence
    from typing import Any, Literal, Protocol

    __all__ = {__all__}


    ''')

    class_def = dedent('''\
    class {name}(Protocol):
        """{docstring}"""
    ''')

    attr_def = with_indent(4)('''\
        @property
        def {name}(self) -> {type}:
            """{docstring}"""
    ''')

    proto_prefix = ""


class DataclassSpec:
    prelude = dedent('''\
    """BIDS validation context dataclasses

    The classes in this module may be used to populate the context for BIDS validation.

    This module has been auto-generated from the BIDS schema version {version}.
    """

    from __future__ import annotations

    import sys
    from dataclasses import dataclass

    TYPE_CHECKING = False
    if TYPE_CHECKING or "sphinx.ext.autodoc" in sys.modules:
        from collections.abc import Mapping, Sequence
        from typing import Any, Literal

        from . import protocols

    if sys.version_info >= (3, 10):
        dc_kwargs = {{"slots": True, "frozen": True}}
    else:  # PY39
        dc_kwargs = {{"frozen": True}}

    __all__ = {__all__}


    ''')

    class_def = dedent('''\
    @dataclass(**dc_kwargs)
    class {name}:
        """{docstring}"""
    ''')

    attr_def = with_indent(4)("""\
        {name}: {type}{default}
        #: {docstring}
    """)

    proto_prefix = "protocols."


def snake_to_pascal(name: str) -> str:
    return "".join(word.capitalize() for word in name.split("_"))


def create_protocol_source(
    class_name: str,
    properties: dict[str, Any],
    metadata: dict[str, Any],
    template: Spec,
    classes: dict[str, str],
) -> str:
    class_name = snake_to_pascal(class_name)
    docstring = metadata.get("description", "").strip()
    if "@property" not in template.attr_def:
        docstring += with_indent(4)("""

            Attributes
            ----------
            """)

    required = metadata.get("required", {})
    optional = [prop for prop in properties if prop not in required]

    lines = []
    for prop_name in (*required, *optional):
        prop_info = properties[prop_name]
        type_, md = typespec_to_source(prop_name, prop_info, template, classes)
        default = ""
        if prop_name not in required:
            type_ = f"{type_} | None"
            default = " = None"
        if not type_.startswith(("int", "float", "str", "bool", "Literal", "Sequence", "Mapping")):
            type_ = f"{template.proto_prefix}{type_}"
        description = md.get("description", "").strip()
        lines.append(
            template.attr_def.format(
                name=prop_name, type=type_, docstring=description, default=default
            )
        )
        # Avoid double-documenting properties
        if "@property" not in template.attr_def:
            docstring += with_indent(4)(f"""\
                {prop_name}: {type_}
                    {description}

                """)

    lines.insert(0, template.class_def.format(name=class_name, docstring=docstring))

    classes[class_name] = "\n".join(lines)
    return class_name


def typespec_to_source(
    name: str,
    typespec: dict[str, Any],
    template: Spec,
    classes: dict[str, str],
) -> tuple[str, dict[str, Any]]:
    """Convert JSON-schema style specification to type and metadata dictionary."""
    tp = typespec.get("type")
    if not tp:
        raise ValueError(f"Invalid typespec: {json.dumps(typespec)}")
    metadata = {
        key: typespec[key] for key in ("name", "description", "required") if key in typespec
    }
    if tp == "object":
        properties = typespec.get("properties")
        if properties:
            type_ = create_protocol_source(
                name, properties=properties, metadata=metadata, template=template, classes=classes
            )
        else:
            type_ = "Mapping[str, Any]"
    elif tp == "array":
        if "items" in typespec:
            subtype, md = typespec_to_source(name, typespec["items"], template, classes=classes)
        else:
            subtype = "Any"
        type_ = f"Sequence[{subtype}]"
    else:
        type_ = {
            "number": "float",
            "string": "str",
            "integer": "int",
        }[tp]
        if type_ == "str" and "enum" in typespec:
            type_ = f"Literal[{', '.join(f'{v!r}' for v in typespec['enum'])}]"
    return type_, metadata


def generate_protocols(
    typespec: dict[str, Any],
    root_class_name: str,
    template: Spec,
) -> dict[str, str]:
    """Generate protocol definitions from a JSON schema typespec."""
    protocols: dict[str, str] = {}
    typespec_to_source(
        root_class_name,
        typespec,
        template=template,
        classes=protocols,
    )
    return protocols


def generate_module(schema: dict[str, Any], class_type: str) -> str:
    """Generate a context module source code from a BIDS schema.

    Returns a tuple containing the module source code and a list of protocol names.
    """
    template: Spec = ProtocolSpec if class_type == "protocol" else DataclassSpec
    protocols = generate_protocols(schema["meta"]["context"], "context", template)
    prelude = template.prelude.format(version=schema["schema_version"], __all__=list(protocols))
    return prelude + "\n\n".join(protocols.values())



from __future__ import annotations

from collections.abc import Mapping
from functools import singledispatch

import pytest
from pyparsing.exceptions import ParseException

from ..expressions import (
    Array,
    ASTNode,
    BinOp,
    Element,
    Function,
    Property,
    RightOp,
    expression,
)
from ..types import Namespace


def test_schema_expressions(schema_obj):
    # Basic smoke test. All of these expressions should parse without error.
    for testexp in schema_obj.meta.expression_tests:
        expression.parse_string(testexp["expression"])


@pytest.mark.parametrize(
    "expr, disambiguated",
    [
        ("a*x**2 + b*x + c == 0", "((((a * (x ** 2)) + (b * x)) + c) == 0)"),
        ("a.b.c[0][1][2].d(3, 4)", "(((a.b).c)[0][1][2].d)(3, 4)"),
        (
            "nifti_header.pixdim[4] == sidecar.RepetitionTime",
            "((nifti_header.pixdim)[4] == (sidecar.RepetitionTime))",
        ),
        (
            """x.y * 10 ** (-3 * index(["a", "b", "c"], w.z))""",
            """((x.y) * (10 ** (-3 * index(['"a"', '"b"', '"c"'], (w.z)))))""",
        ),
        ("a && b && c || d || e && f", "((a && (b && c)) || (d || (e && f)))"),
        ("! a && b || ! c == d || e", "(((!a) && b) || ((!(c == d)) || e))"),
    ],
)
def test_expression_associations(expr, disambiguated):
    # The goal of this test is to use the str() function on the AST to
    # reproduce the original expression, but with parentheses demonstrating
    # associativity. For example `a.b.c` could be parsed `(a.b).c` or `a.(b.c)`.
    parse_results = expression.parse_string(expr)
    assert len(parse_results) == 1
    ast = parse_results[0]
    assert str(ast) == disambiguated


def test_selectors(schema_obj):
    # Selectors happen at the second level in a file
    # For directories of files, need to look three down
    for rules, level in (
        (schema_obj.rules.checks, 3),
        (schema_obj.rules.dataset_metadata, 2),
        (schema_obj.rules.sidecars, 3),
        (schema_obj.rules.tabular_data, 3),
        (schema_obj.rules.errors, 2),
    ):
        keys = (key for key in rules.keys(level=level) if key.endswith("selectors"))
        for key in keys:
            for selector in rules[key]:
                ast = expression.parse_string(selector)[0]
                assert isinstance(ast, ASTNode)


def test_checks(schema_obj):
    checks = (
        expr
        for key, value in schema_obj.rules.checks.items(level=3)
        if key.endswith("checks")
        for expr in value
    )
    for check in checks:
        ast = expression.parse_string(check)[0]
        assert isinstance(ast, ASTNode)


@pytest.mark.parametrize(
    "expr",
    (
        "func(x, y",
        "lhs == ",
        "assignment = not.a.thing",
    ),
)
def test_expected_failures(expr):
    with pytest.raises(ParseException):
        expression.parse_string(expr)


def walk_schema(schema_obj, predicate):
    for key, value in schema_obj.items():
        if predicate(key, value):
            yield key, value
        if isinstance(value, Mapping):
            for subkey, value in walk_schema(value, predicate):
                yield f"{key}.{subkey}", value


def test_valid_sidecar_field(schema_obj):
    """Check sidecar fields actually exist in the metadata listed in the schema.

    Test failures are usually due to typos.
    """
    field_names = {field.name for key, field in schema_obj.objects.metadata.items()}

    for key, rule in walk_schema(
        schema_obj.rules, lambda k, v: isinstance(v, Mapping) and v.get("selectors")
    ):
        for selector in rule["selectors"]:
            ast = expression.parse_string(selector)[0]
            for name in find_names(ast):
                if name.startswith(("json.", "sidecar.")):
                    assert name.split(".", 1)[1] in field_names, (
                        f"Bad field in selector: {name} ({key})"
                    )
        for check in rule.get("checks", []):
            ast = expression.parse_string(check)[0]
            for name in find_names(ast):
                if name.startswith(("json.", "sidecar.")):
                    assert name.split(".", 1)[1] in field_names, (
                        f"Bad field in check: {name} ({key})"
                    )


def test_test_valid_sidecar_field():
    schema_obj = Namespace.build(
        {
            "objects": {
                "metadata": {
                    "a": {"name": "a"},
                }
            },
            "rules": {"myruleA": {"selectors": ["sidecar.a"], "checks": ["json.a == sidecar.a"]}},
        }
    )
    test_valid_sidecar_field(schema_obj)

    schema_obj.objects.metadata.a["name"] = "b"
    with pytest.raises(AssertionError):
        test_valid_sidecar_field(schema_obj)


@singledispatch
def find_names(node: ASTNode | str):
    # Walk AST nodes
    if isinstance(node, BinOp):
        yield from find_names(node.lh)
        yield from find_names(node.rh)
    elif isinstance(node, RightOp):
        yield from find_names(node.rh)
    elif isinstance(node, Array):
        for element in node.elements:
            yield from find_names(element)
    elif isinstance(node, Element):
        yield from find_names(node.name)
        yield from find_names(node.index)
    elif isinstance(node, (int, float)):
        return
    else:
        raise TypeError(f"Unexpected node type: {node!r}")


@find_names.register
def find_function_names(node: Function):
    yield node.name
    for arg in node.args:
        yield from find_names(arg)


@find_names.register
def find_property_name(node: Property):
    # Properties are left-associative, so expand the left side
    yield f"{next(find_names(node.name))}.{node.field}"


@find_names.register
def find_identifiers(node: str):
    if not node.startswith(('"', "'")):
        yield node



"""Parsing utilities for BIDS Schema expression language"""

from functools import partial

from pyparsing import (
    Forward,
    Literal,
    Optional,
    StringEnd,
    StringStart,
    Suppress,
    common,
    delimited_list,
    one_of,
    quoted_string,
)


def parse(expression_string: str) -> "ASTNode":
    """Convert a BIDS schema expression into an abstract syntax tree

    EBNF-ish grammar::

        # In order of binding (loosest to tightest)
        orOp  :: '||'
        andOp :: '&&'
        notOp :: '!'
        cmpOp :: '==' | '!=' | '<' | '<=' | '>' | '>=' | 'in'
        addOp :: '+' | '-'
        mulOp :: '*' | '/' | '%'
        expOp :: '**'

        item    :: '(' test ')' | '[' [testList] '] | '{' '}' | NAME | NUMBER | STRING
        trailer :: '(' [testList] ') | '[' test ']' | '.' NAME
        atom    :: item trailer*

        factor :: atom [ expOp factor ]*
        term   :: factor [ mulOp factor ]*
        expr   :: term [ addOp term ]*

        comparison :: expr [ compOp expr ]*
        notTest    :: not notTest | comparison
        andTest    :: notTest [ '&&' andTest ]*
        test       :: andTest [ '||' test ]*

        testList   :: test [ ',' test ]*

        expression :: ^ test $
    """
    # Alternative: (possibly with 'null' as additional literal)
    # This makes things like '1(a, b, c)' or '"string"[0]' syntax errors
    #
    # literal :: NUMBER | STRING
    # item    :: '(' test ')' | '[' testList '] | NAME
    # atom    :: literal | item trailer*

    return expression.parse_string(expression_string)[0]


orOp = Literal("||")
andOp = Literal("&&")
notOp = Literal("!")
compOp = one_of(("==", "!=", "<", "<=", ">", ">=", "in"))
addOp = one_of(("+", "-"))
mulOp = one_of(("*", "/"))
expOp = Literal("**")

lpar, rpar = Suppress("("), Suppress(")")
lsqr, rsqr = Suppress("["), Suppress("]")
dot = Suppress(".")

# Right-associative expressions need to be recursively defined
factor = Forward()
notTest = Forward()
andTest = Forward()
test = Forward()

testlist = delimited_list(test)

# Numbers and strings are base types, this could be expanded with bools and null
# if it seems useful
literal = common.number | quoted_string

# Items are units that operations can be done on, including arithmetic, comparison,
# function calls, index lookups and attribute lookups
array = lsqr + Optional(testlist) + rsqr
parenthetical = lpar + test + rpar
# If object literals ever occur in real expressions, we'll need to define this
obj_literal = Literal("{}")
item = parenthetical | array | obj_literal | common.identifier | literal

# Trailers are function calls, array indexes, and object attributes
function_call = lpar + Optional(testlist) + rpar
array_lookup = lsqr + test + rsqr
object_lookup = dot + common.identifier
trailer = function_call | array_lookup | object_lookup

# An atom might have some lookups done, but now it can be part of an arithmetic
# expression
atom = item + (trailer)[...]

# Arithmetic expressions
factor <<= atom + (expOp + factor)[...]  # Right-associative
term = factor + (mulOp + factor)[...]
expr = term + (addOp + term)[...]

# Logic expressions (tests, to avoid name collision)
comparison = expr + (compOp + expr)[...]
notTest <<= notOp + notTest | comparison
andTest <<= notTest + (andOp + andTest)[...]  # Right-associative
test <<= andTest + (orOp + test)[...]  # Right-associative

# Schema expressions must parse from start to finish
expression = StringStart() + test + StringEnd()


# With the grammar defined, we need the following syntax elements
class ASTNode:
    """AST superclass

    Defines basic repr. Subclasses should define __str__ with enough
    parentheses to make associations unambiguous.
    """

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.__dict__}>"

    def __str__(self) -> str:
        raise NotImplementedError


class RightOp(ASTNode):
    """Right-associative unary operator"""

    def __init__(self, tokens):
        self.op, self.rh = tokens

    def __str__(self):
        return f"({self.op}{self.rh})"

    @classmethod
    def maybe(cls, tokens):
        """Construct class only if not degenerate"""
        if len(tokens) < 2:
            return tokens
        return cls(tokens)


class BinOp(ASTNode):
    """Binary operator"""

    def __init__(self, tokens):
        self.lh, self.op, self.rh = tokens

    @classmethod
    def maybe(cls, tokens):
        """Construct class if not degenerate

        * Right-associative: ``outer <<= inner + (op + outer)[...]``
        * Left-associative: ``outer = inner + (op + inner)[...]``

        In the right-associative case, we get ``[inner, op, BinOp(outer)]``,
        so we loop once and are done.

        In the left-associative case, we get ``[inner, op, inner, op, ...]``,
        convert to ``[BinOp(inner, op, inner), op, inner, ...]``
        """
        while len(tokens) >= 3:
            tokens = [cls(tokens[:3])] + tokens[3:]
        return tokens

    def __str__(self):
        return f"({self.lh} {self.op} {self.rh})"


class Function(ASTNode):
    """Function call"""

    def __init__(self, name, args):
        self.name = name
        self.args = args

    def __str__(self):
        arglist = ", ".join(map(str, self.args))
        return f"{self.name}({arglist})"


class Element(ASTNode):
    """Array element lookup"""

    def __init__(self, name, index):
        self.name = name
        self.index = index

    def __str__(self):
        return f"{self.name}[{self.index}]"


class Property(ASTNode):
    """Object property lookup"""

    def __init__(self, name, field):
        self.name = name
        self.field = field

    def __str__(self):
        return f"({self.name}.{self.field})"


class Array(ASTNode):
    """Array literal"""

    def __init__(self, tokens):
        self.elements = list(tokens)

    def __str__(self):
        return str(self.elements)


class Object(ASTNode):
    """Object literal, unused outside tests, so degenerate"""

    def __init__(self, tokens):
        pass

    def __str__(self):
        return "{}"


array.set_parse_action(Array)
obj_literal.set_parse_action(Object)

# Function calls and item lookups need to be constructed partially
function_call.set_parse_action(lambda t: partial(Function, args=list(t)))
array_lookup.set_parse_action(lambda t: partial(Element, index=t[0]))
object_lookup.set_parse_action(lambda t: partial(Property, field=t[0]))


# Once the atom is complete, we can build the left-associative tree by completing application
def atomize(tokens):
    item = tokens.pop(0)
    for trailer in tokens:
        item = trailer(item)
    return item


atom.set_parse_action(atomize)

# Arithmetic expressions can all be degenerate, so use maybe to pass through
factor.set_parse_action(BinOp.maybe)
term.set_parse_action(BinOp.maybe)
expr.set_parse_action(BinOp.maybe)

comparison.set_parse_action(BinOp.maybe)
notTest.set_parse_action(RightOp.maybe)
andTest.set_parse_action(BinOp.maybe)
test.set_parse_action(BinOp.maybe)



"""Functions for rendering elements of the schema in the specification text."""

from bidsschematools.render.tables import (
    make_columns_table,
    make_entity_table,
    make_json_table,
    make_metadata_table,
    make_sidecar_table,
    make_subobject_table,
    make_suffix_table,
)
from bidsschematools.render.text import (
    define_allowed_top_directories,
    define_common_principles,
    make_entity_definitions,
    make_filename_template,
    make_glossary,
    render_text,
)

__all__ = [
    "make_entity_table",
    "make_suffix_table",
    "make_json_table",
    "make_sidecar_table",
    "make_metadata_table",
    "make_subobject_table",
    "make_columns_table",
    "make_entity_definitions",
    "make_glossary",
    "make_filename_template",
    "define_common_principles",
    "define_allowed_top_directories",
    "render_text",
]



import os
import shutil

import pytest


def require_env(var):
    env = os.environ.get(var)
    return pytest.mark.skipif(
        not env, reason=f"To activate this test/feature `export {var}=1` and re-run."
    )


@require_env("BIDSSCHEMATOOLS_RELEASE")
def test_make_archive(tests_data_dir, bids_examples, bids_error_examples):
    """
    ATTENTION! This is not a test!
    Create static testdata archive containing the bidsschematools data reference whitelist.

    Notes
    -----
    Due to intricacies arising from:

    (1) fixtures not working outside of pytest
    (2) implicit teardown leveraging tempdata removal (while held open by yield)
    (3) wrappers evaluating the yield statement
    (4) the desire to not download testdata twice for archive creation

    testdata archive creation is now inconspicuously posing as a test.
    """

    ignore_git = shutil.ignore_patterns(".git*")
    target_examples = tests_data_dir / "bids-examples"
    target_error_examples = tests_data_dir / "bids-error-examples"
    if bids_examples != target_examples:
        shutil.copytree(bids_examples, target_examples, ignore=ignore_git)
    if bids_error_examples != target_error_examples:
        shutil.copytree(bids_error_examples, target_error_examples, ignore=ignore_git)
